// FLASH Programmer:
//   This code programs the AT29C256 Flash ROM
//   The "Flash" data is in the source code.
//   Note: The standard Arduino Nano does not have enough free
//     memory (Flash) to program the whole Flash ROM in one pass!
//
//   Programmer connections:
//     D0  <-> TX
//     D1  <-> RX
//     D2  <-> Flash D0
//     D3  <-> Flash D1
//     D4  <-> Flash D2
//     D5  <-> Flash D3
//     D6  <-> Flash D4
//     D7  <-> Flash D5
//     D8  <-> Flash D6
//     D9  <-> Flash D7
//     THE LOW ORDER ADDRESS FOR THE PROGRAMMING PAGE (64 BYTES)
//     A0  <-> Flash A0
//     A1  <-> Flash A1
//     A2  <-> Flash A2
//     A3  <-> Flash A3
//     A4  <-> Flash A4
//     A5  <-> Flash A5
//     THE HIGH ORDER ADDRESS (PAGES) ARE CLOCK (>) FROM THE D0-7 USING D13 (LED)
//     >D2 <-> Flash A6
//     >D3 <-> Flash A7
//     >D4 <-> Flash A8
//     >D5 <-> Flash A9
//     >D6 <-> Flash A10
//     >D7 <-> Flash A11
//     >D8 <-> Flash A12
//     >D9 <-> Flash A13
//     D10 <-> Flash D14 (THIS IS SET AND FORGET)
//     D11 <-> Flash !OE
//     D12 <-> Flash !WE
//     D13 <-> CLOCK (RISING EDGE) THE HIGH ADDRESS

// Here is the "FLASH" data
#include <avr/pgmspace.h>
const int flashBytes=64;
/*
const byte PROGMEM FLASH[flashBytes] = {
  0X00, 0X01, 0X02, 0X03,
  0X04, 0X05, 0X06, 0X07,
  0X08, 0X09, 0X0A, 0X0B,
  0X0C, 0X0D, 0X0E, 0X0F,
  0XF0, 0XF1, 0XF2, 0XF3,
  0XF4, 0XF5, 0XF6, 0XF7,
  0XF8, 0XF9, 0XFA, 0XFB,
  0XFC, 0XFD, 0XFE, 0XFF,
  0X00, 0X01, 0X20, 0X30,
  0X40, 0X50, 0X60, 0X70,
  0X80, 0X90, 0XA0, 0XB0,
  0XC0, 0XD0, 0XE0, 0XF0,
  0X0F, 0X1F, 0X2F, 0X3F,
  0X4F, 0X5F, 0X6F, 0X7F,
  0X8F, 0X9F, 0XAF, 0XBF,
  0XCF, 0XDF, 0XEF, 0XFF
};
*/
const byte PROGMEM FLASH[flashBytes] = {
  0X00, 0X00, 0X00, 0X00,
  0Xff, 0Xff, 0Xff, 0Xff,
  0X00, 0X00, 0X00, 0X00,
  0Xff, 0Xff, 0Xff, 0Xff,
  0X00, 0X00, 0X00, 0X00,
  0Xff, 0Xff, 0Xff, 0Xff,
  0X00, 0X00, 0X00, 0X00,
  0Xff, 0Xff, 0Xff, 0Xff,
  0X00, 0X00, 0X00, 0X00,
  0Xff, 0Xff, 0Xff, 0Xff,
  0X00, 0X00, 0X00, 0X00,
  0Xff, 0Xff, 0Xff, 0Xff,
  0X00, 0X00, 0X00, 0X00,
  0Xff, 0Xff, 0Xff, 0Xff,
  0X00, 0X00, 0X00, 0X00,
  0Xff, 0Xff, 0Xff, 0Xff
};
void SetA14Low(void)
{
  // PortB XXC!!ADD     
  //       XXLWO176
  //       XXKEE4
  PORTB&= B11111011;  
}

void SetA14High(void)
{
  // PortB XXC!!ADD     
  //       XXLWO176
  //       XXKEE4
  PORTB|= B00000100;    
}

void WriteData(byte Data)
{
  // PortB XXC!!ADD     
  //       XXLWO176
  //       XXKEE4
  // PortD DDDDDDRT
  //       543210XX
  
  /* Output Enable unset */
  PORTB|= B00001000;  
  /* Set Data */
  PORTD=(PORTD&B00000011)|((Data<<2)&B11111100);
  PORTB=(PORTB&B11111100)|((Data>>6)&B00000011);
  /* Set D0-7 to output */
  DDRD |= B11111100;
  DDRB |= B00000011;
  /* Write Data */  
  delayMicroseconds(2); // Actually this is only ~1 us
  PORTB&= B11101111;
  delayMicroseconds(2); // Actually this is only ~1 us
  PORTB|= B00010000;
  /* Set D0-7 as input */
  DDRD &= B00000011;
  DDRB &= B11111100;  
}

byte ReadData(void)
{
  byte Data;
  /* Pin Mapping */
  // PortB XXC!!ADD     
  //       XXLWO176
  //       XXKEE4
  // PortD DDDDDDRT
  //       543210XX
  
  /* Output Enable unset */
  PORTB|= B00001000;
  /* Set D0-7 as input */
  DDRD &= B00000011;
  DDRB &= B11111100;
  /* Clear D0-7 */
  PORTD&= B00000011;
  PORTB&= B11111100;
  /* Output Enable set */
  PORTB&= B11110111;
  /* Read the D0-7 */
  delayMicroseconds(2); // Actually this is only ~1 us
  Data=((PIND>>2)&B00111111)|((PINB<<6)&B11000000);
  /* Output Enable unset */
  PORTB|= B00001000;

  return Data;
}

void SetData(byte Data)
{
  // PortB XXC!!ADD     
  //       XXLWO176
  //       XXKEE4
  // PortD DDDDDDRT
  //       543210XX
  
  /* Output Enable unset */
  PORTB|= B00001000;  
  /* Set Data */
  PORTD=(PORTD&B00000011)|((Data<<2)&B11111100);
  PORTB=(PORTB&B11111100)|((Data>>6)&B00000011);
  /* Set D0-7 to output */
  DDRD |= B11111100;
  DDRB |= B00000011;
}
void SetAddr(unsigned int Addr)
{
  byte temp;
  /* Pin Mapping */
  // PortB XXC!!ADD     
  //       XXLWO176
  //       XXKEE4
  // PortC XXAAAAAA
  //       XX543210
  // PortD DDDDDDRT
  //       543210XX
  
  /* Output Enable unset */
  PORTB|= B00001000;
  /* Set A0-5 */
  PORTC=(PORTC&B11000000)|(Addr&B00111111);
  /* Set Address A6-A13 to D0-7 */
  temp=(byte)(Addr>>6);
  PORTD=(PORTD&B00000011)|((temp<<2)&B11111100);
  PORTB=(PORTB&B11111100)|((temp>>6)&B00000011);
  /* Set D0-7 to output */
  DDRD |= B11111100;
  DDRB |= B00000011;
  /* Clock D0-7 into A6-A13 */
  PORTB&= B11011111;
  delayMicroseconds(2); // Actually this is only ~1 us
  PORTB|= B00100000;
  /* Clear D0-7 */
  PORTD&= B00000011;
  PORTB&= B11111100;
  /* Set D0-7 as input */
  DDRD &= B00000011;
  DDRB &= B11111100;
}

void ReadFLASH(unsigned int firstPage,unsigned int lastPage)
{
  byte Data;
  unsigned int Addr;

  for (Addr=firstPage*64;Addr<(lastPage+1)*64;Addr++) {
    SetAddr(Addr);
    Data=ReadData();
    if (Addr%4==0) {
      Serial.println();
      if (Addr<=0x000f) {
        Serial.print("000");
      } else if (Addr<=0x00ff) {
        Serial.print("00");
      } else if (Addr<=0x0fff) {
        Serial.print("0");
      } 
      Serial.print(Addr,HEX);
      Serial.print(":");
    }
    Serial.print(" ");
    if (Data<=0x0f) Serial.print("0");
    Serial.print(Data,HEX);
    delay(10); // Enough time for 8 characters per loop at 9600 baud
  }
}

void VerifyFLASH(unsigned int firstPage,unsigned int lastPage)
{
  byte Data;
  unsigned int Addr;

  for (Addr=firstPage*64;Addr<(lastPage+1)*64;Addr++) {
    SetAddr(Addr);
    if (Addr%4==0) {
      Serial.println();
      if (Addr<=0x000f) {
        Serial.print("000");
      } else if (Addr<=0x00ff) {
        Serial.print("00");
      } else if (Addr<=0x0fff) {
        Serial.print("0");
      } 
      Serial.print(Addr,HEX);
      Serial.print(":");
    }
    Serial.print(" ");
    Data=pgm_read_byte(FLASH+Addr);
    if (Data<=0x0f) Serial.print("0");
    Serial.print(Data,HEX);
    Serial.print("|");
    Data=ReadData();
    if (Data<=0x0f) Serial.print("0");
    Serial.print(Data,HEX);
    delay(15); // Enough time for 12 characters per loop at 9600 baud
  }
}

void WriteFLASH(unsigned int firstPage,unsigned int lastPage)
{
  byte Data;
  unsigned int Page;
  unsigned int AddrHigh;
  unsigned int AddrLow;

  for (Page=firstPage;Page<=lastPage;Page++) {
    AddrHigh=(Page<<6);
    SetAddr(AddrHigh);
    // Must be less than 150us per cycle!
    for (AddrLow=0;AddrLow<64;AddrLow++) {
      /* Set A0-5 */
      PORTC=(PORTC&B11000000)|(AddrLow&B00111111);
      Data=pgm_read_byte(FLASH+AddrHigh+AddrLow);
      WriteData(Data);
    }
    delay(20); // Maximum write cycle time is 10 ms
  }
  SetAddr(0);
}

void setup()
{
  /* INITIALISE PORTS */
  // PortC XXAAAAAA
  //       XX543210  
  DDRC |= B00111111; // Set A0-5 outputs
  PORTC|= B00000000; // Set A0-5 low
  
  // PortB XXC!!ADD 
  //       XXLWO176
  //       XXKEE4
  PORTB|= B00111000; // Set CLK, !WE & !OE high
  PORTB&= B11111000; // Set D14, D7 & D6 LOW
  DDRB &= B11111100; // Set D7 & D6 as inputs
  DDRB |= B00111100; // Set Set CLK, !WE, !OE & D14 as outputs

  // PortD DDDDDDRT
  //       543210XX
  DDRD &= B00000011; // Set D0-5 as inputs
  PORTD&= B00000011; // Set D0-5 low

  // Set Serial
  Serial.begin(9600);

  // Set inital address (0x0000)
  SetA14Low();
  SetAddr(0);
}

void loop()
{
  char Ans;
  
  Serial.print("");
  delay(100);
  Serial.println("Welcome to the AT19C256 FLASH Programmer V1.0");

  /* First Read FLASH */
  delay(100);                                     // Wait for output buffer to clear
  Serial.print("First read the FLASH (y/n)?");    // Message
  while (Serial.available()) Serial.read();       // Flush input buffer
  while (!Serial.available());                    // Wait for response
  Ans=Serial.read();                              // Get answer
  if ((Ans=='y')||(Ans=='Y')) {
    // Read FLASH
    ReadFLASH(0,0); // Page 0 
    delay(100);
    Serial.println();
    Serial.println("FLASH read done.");
  } else {
    delay(100);
    Serial.println();
    Serial.println("FLASH read aborted.");
  }
  Serial.println();
      
  // Write FLASH
  delay(100);                                     // Wait for output buffer to clear
  Serial.print("Write the FLASH (y/n)?");         // Message
  while (Serial.available()) Serial.read();       // Flush input buffer
  while (!Serial.available());                    // Wait for response
  Ans=Serial.read();                              // Get answer
  if ((Ans=='y')||(Ans=='Y')) {
    // Write FLASH
    WriteFLASH(0,0); // Page 0
    delay(100);
    Serial.println();
    Serial.println("FLASH write done.");
  } else {
    delay(100);
    Serial.println();
    Serial.println("FLASH write aborted.");
  }
  Serial.println();
  
  // Verify FLASH
  delay(100);                                     // Wait for output buffer to clear
  Serial.print("Verify (read) the FLASH (y/n)?"); // Message
  while (Serial.available()) Serial.read();       // Flush input buffer
  while (!Serial.available());                    // Wait for response
  Ans=Serial.read();                              // Get answer
  if ((Ans=='y')||(Ans=='Y')) {
    // Verify FLASH
    VerifyFLASH(0,0); // Page 0 
    delay(100);
    Serial.println();
    Serial.println("FLASH verify done.");
  } else {
    delay(100);
    Serial.println();
    Serial.println("FLASH verify aborted.");
  }
  
  /* Done Message */
  delay(100);
  Serial.println();
  Serial.println("FLASH Programmer done.");

  while (true);

}

